/* Just a simple example program that creates a terminal in a frame
 * and lets the user interact with it.
 *
 * To compile:
 *    gcc -o boxshell boxshell.c $(pkg-config madtty --cflags --libs)
 */

#include <ncurses.h>

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <termios.h>
#include <time.h>

#include <madtty/madtty.h>

static int getout = 0, sigwinch = 0;
static int screen_w, screen_h;
static WINDOW *term_win;
static struct timeval const slice = { 0, 1000 * 1000 / 100 };

void handler(int signo)
{
    switch (signo) {
      case SIGCHLD:
        getout = 1;
        break;
      case SIGWINCH:
        sigwinch = 1;
        break;
    }
}

static struct timeval timeval_add(struct timeval a, struct timeval b)
{
    int usec = a.tv_usec + b.tv_usec;
    a.tv_sec += b.tv_sec;
    while (usec > 1000 * 1000) {
        a.tv_sec += 1;
        usec -= 1000 * 1000;
    }
    a.tv_usec = usec;
    return a;
}

static int is_expired(struct timeval now, struct timeval expiry)
{
    return now.tv_sec > expiry.tv_sec
        || (now.tv_sec == expiry.tv_sec && now.tv_usec > expiry.tv_usec);
}

int main(void)
{
    madtty_t *rt;
    int dirty = 0, pty;
    struct timeval next;

    signal(SIGCHLD, handler);
    signal(SIGWINCH, handler);

    madtty_initialize();
    getmaxyx(stdscr, screen_h, screen_w);

    /* create a window with a frame */
    term_win = newwin(screen_h - 2, screen_w - 2, 1, 1);
    rt = madtty_create(screen_h - 2, screen_w -2);
    {
        const char *path = getenv("SHELL") ?: "/bin/sh";
        const char *args[] = { path, "--login", NULL};

        madtty_forkpty(rt, path, args, &pty);
    }

    /* keep reading keypresses from the user and passing them to the terminal;
     * also, redraw the terminal to the window at each iteration */
    gettimeofday(&next, NULL);
    while (!getout) {
        struct timeval tv = { 0, 1000 * 1000 / 100 };
        fd_set rfds;
        int ch;

        FD_ZERO(&rfds);
        FD_SET(0, &rfds);
        FD_SET(pty, &rfds);

        if (select(pty + 1, &rfds, NULL, NULL, &tv) > 0) {
            if (FD_ISSET(pty, &rfds)) {
                madtty_process(rt);
                dirty = 1;
            }
        }

        if (sigwinch) {
            int fd, cols = -1, rows = -1;
            struct winsize w;

            if ((fd = open("/dev/tty", O_RDONLY)) != -1) {
                if (ioctl(fd, TIOCGWINSZ, &w) != -1) {
                    rows = w.ws_row;
                    cols = w.ws_col;
                }
                close(fd);
            }
            if (rows <= 0) {
                rows = atoi(getenv("LINES") ?: "24");
            }
            if (cols <= 0) {
                cols = atoi(getenv("COLUMNS") ?: "80");
            }

            resizeterm(rows, cols);
            madtty_resize(rt, rows - 2, cols - 2);
            wresize(term_win, rows - 2, cols - 2);
            sigwinch = 0;
            erase();
        }

        while ((ch = getch()) != ERR) {
            madtty_keypress(rt, ch); /* pass the keypress for handling */
            dirty = 1;
        }

        gettimeofday(&tv, NULL);
        if (dirty && is_expired(tv, next)) {
            madtty_draw(rt, term_win, 0, 0);
            wrefresh(term_win);
            dirty = 0;
            next = timeval_add(tv, slice);
        }
    }

    endwin();
    return 0;
}
